import yaml
import os
import subprocess


class VerificationTarget:
    def __init__(self):
        self.core_file = None
        self.core_name = None
        self.target_name = None
        self.eda_tool = None
        self.outpath = None,
        # Attributes for verification command.
        self.command = 'fusesoc'
        self.command_arguments = ['--cores-root', '.', 'run', '--target']
        # Attributes for verification results.
        self.passed = False
        self.number_of_errors = 0
        self.number_of_warnings = 0
        # TODO: add attribute with time for verification

    def _prepare_output_directory(self, outpath):
        core_name = self.core_name.strip(':').replace(':', '-')

        self.outpath = outpath + '/' + core_name + '/' + self.target_name + '/'
        try:
            os.makedirs(self.outpath)
        except FileExistsError:
            pass

    def _prepare_analyze_options(self):
        self.command_arguments.append(self.target_name)
        self.command_arguments.append(self.core_name)

        # Add path for pre analyzed libraries for GHDL.
        #  Right now it is hardcoded to the default GHDL installation path.
        #  If there is a need correct path will be detected automatically.
        if self.eda_tool == "ghdl":
            self.command_arguments.append('--analyze_options')
            self.command_arguments.append("\\-P/usr/local/lib/ghdl/vendors -frelaxed-rules -fpsl")

    def _prepare_run_options(self):
        if self.eda_tool == "ghdl":
            ghdl_psl_report_file = self.outpath + "ghdl_psl_report.json"
            ghdl_vcd_file = self.outpath + "ghdl.ghw"
            self.command_arguments.append('--run_options')
            self.command_arguments.append("\\--psl-report=" + ghdl_psl_report_file + " --wave=" + ghdl_vcd_file)

    def _prepare_for_verification(self, outpath):
        self._prepare_output_directory(outpath)
        self._prepare_analyze_options()
        self._prepare_run_options()

    def _verify(self):
        print("Verifying core: " + self.core_name + ", target: " + self.target_name)

        output = subprocess.run([self.command] + self.command_arguments, capture_output=True, encoding='utf-8')

        if output.returncode == 0:
            self.passed = True

        self.number_of_errors = output.stdout.lower().count('error')
        self.number_of_warnings = output.stdout.lower().count('warn')

        return output

    def verify_to_file(self, outpath):
        self._prepare_for_verification(outpath)

        output = self._verify()

        with open(self.outpath + 'output.txt', 'w') as f:
            f.write("***** STDERR *****\n\n")
            f.write(output.stderr)

            f.write("\n***** STDOUT *****\n\n")
            f.write(output.stdout)

    def verify_to_console(self, outpath):
        self._prepare_for_verification(outpath)

        output = self._verify()

        print("\n***** STDERR *****")
        print(output.stderr)

        print("***** STDOUT *****")
        print(output.stdout)

        print("core: " + self.core_name + ", target: " + self.target_name)

        if self.passed:
            print("PASSED")
        else:
            print("FAILED")

        print("ERRORS:    " + str(self.number_of_errors))
        print("WARNINGS:  " + str(self.number_of_warnings))
        print('\n')


def from_file(file):
    """
    Returns verification targets from single FuseSoc .core file.

    :param file: .core file
    :return: list with verification targets
    """
    with open(file) as f:
        try:
            yaml_dict = yaml.safe_load(f)
        except yaml.YAMLError as exc:
            print(exc)

    verification_targets = []
    for target in yaml_dict['targets']:
        if target.startswith('tb_') or target.endswith('_tb'):
            ver_target = VerificationTarget()
            ver_target.core_file = f.name
            ver_target.core_name = yaml_dict['name']
            ver_target.target_name = target
            ver_target.eda_tool = yaml_dict['targets'][target]['default_tool']
            verification_targets.append(ver_target)

    return verification_targets



import os
import argparse
import subprocess
import pathlib
from datetime import datetime
import zipfile
import shutil

from fsva import verification_target


VERSION = '0.0.0'


def parse_command_line_arguments():
    parser = argparse.ArgumentParser(
        prog="FuseSoc Verification Automation",
        description="Tool for automating verification process for HDL projects using FuseSoc build tool."
    )

    # Positional arguments for single core runs and console output
    parser.add_argument("--version", help="Display fsva version.", action='version', version=VERSION)
    parser.add_argument("core", nargs='?', help="Core to verify - console output.")
    parser.add_argument("target", nargs='?', help="Verification target to run - console output.")

    parser.add_argument('-w', '--workpath', default='.',
                        help="Work path. Path to recursively look for FuseSoc .core files.")
    parser.add_argument('-o', '--outdir', default='_fsva',
                        help="Output directory name. This directory is created in workpath.")
    parser.add_argument('-c', '--compress', action='store_true',
                        help="Automatically compress output directory if all tests pass.")
    parser.add_argument('-n', '--numprocesses', default=1,
                        help="Number of processes allowed to be spawn in parallel")

    return parser.parse_args()


def check_fusesoc_installed():
    try:
        out = subprocess.check_output(['fusesoc', '--version'])
        print("Found FuseSoc version: " + out.decode('utf-8'))
    except:
        print("FuseSoc not found!")
        exit(1)


def check_workpath_exists(path):
    if not os.path.exists(path):
        print("Workpath does not exist!")
        exit(1)


def verify_single_core(verification_targets, core, target, outpath):
    ver_targets = []
    for t in verification_targets:
        if t.core_name == core:
            if target is None or t.target_name == target:
                ver_targets.append(t)

    if not ver_targets:
        print("No verification targets found for given core: " + core + ", target: " + target)

    for t in ver_targets:
        t.verify_to_console(outpath)


def print_summary(file, msg):
    print(msg, end='')
    file.write(msg)


def summarize(verification_targets, outpath, start_time):
    num_targets = len(verification_targets)
    num_passed = sum(1 for x in verification_targets if x.passed)
    num_failed = num_targets - num_passed
    all_passed = True

    print()

    with open(outpath + '/summary', 'w') as f:
        num_errors = 0
        num_warnings = 0

        for target in verification_targets:
            if target.passed:
                print_summary(f, "PASSED: core: " + target.core_name + ", target: " + target.target_name + "\n")
            else:
                all_passed = False
                print_summary(f, "FAILED: core: " + target.core_name + ", target: " + target.target_name + "\n")

            if target.number_of_errors > 0:
                num_errors += target.number_of_errors

                print_summary(f, "ERRORS (" + str(target.number_of_errors) + "): core: " + target.core_name + ", target: " + target.target_name + "\n")

            if target.number_of_warnings > 0:
                num_warnings += target.number_of_warnings

                print_summary(f, "WARNINGS (" + str(target.number_of_warnings) + "): core: " + target.core_name + ", target: " + target.target_name + "}\n")

            if not target.passed or target.number_of_warnings > 0:
                print_summary(f, "For more details check directory:" + target.outpath + "\n")

            print_summary(f, "\n")

        s = (datetime.now() - start_time).seconds
        hours, remainder = divmod(s, 3600)
        minutes, seconds = divmod(remainder, 60)

        summary = """
VERIFICATION SUMMARY:
  Total verification time: {} h {} min {} s
  TARGETS:  {}
  PASSED:   {} ({:.2%})
  FAILED:   {} ({:.2%})
  ERRORS:   {}
  WARNINGS: {}\n
""".format(hours, minutes, seconds,
           num_targets,
           num_passed, num_passed/num_targets,
           num_failed, num_failed/num_targets,
           num_errors,
           num_warnings)

        print_summary(f, summary)

    if not all_passed:
        print("At least one verification target failed. Check summary for details.")
        exit(1)


def compress_output_directory(outpath):
    print(outpath)

    dirs_names = outpath.split('/')

    os.chdir(dirs_names[-2])
    zipf = zipfile.ZipFile(dirs_names[-1] + '.zip', 'w', zipfile.ZIP_DEFLATED)

    for root, dirs, files in os.walk(outpath):
        for file in files:
            root_dirs = root.split('/')
            idx = root_dirs.index(dirs_names[-1])
            zip_dirs = root_dirs[idx:]
            arcname = '/'.join(zip_dirs) + '/' + file
            zipf.write(os.path.join(root, file), arcname)

    zipf.close()
    shutil.rmtree(dirs_names[-1])


def main():
    cmd_line_args = parse_command_line_arguments()

    if int(cmd_line_args.numprocesses) > 1:
        print("Parallel verification is not yet supported")
        exit(1)

    check_fusesoc_installed()

    check_workpath_exists(cmd_line_args.workpath)
    workpath = os.path.abspath(cmd_line_args.workpath)
    os.chdir(workpath)

    start_time = datetime.now()

    core_files = []
    for file in pathlib.Path(workpath).glob('**/*.core'):
        core_files.append(file)

    if not core_files:
        print("No .core files found under the workpath")
        exit(1)

    verification_targets = []
    for file in core_files:
        verification_targets += verification_target.from_file(file)

    if not verification_targets:
        print("No verification targets found in .core files")
        exit(1)

    outpath = workpath + '/' + cmd_line_args.outdir + '/'

    if cmd_line_args.core:
        outpath += 'tmp'
        verify_single_core(verification_targets, cmd_line_args.core, cmd_line_args.target, outpath)
        exit(0)

    outpath += datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

    for target in verification_targets:
        target.verify_to_file(outpath)

    summarize(verification_targets, outpath, start_time)

    if cmd_line_args.compress:
        compress_output_directory(outpath)


if __name__ == '__main__':
    main()
